"""Test case of a grid that intersects a shape."""
import shutil
import numpy as np
import geopandas as gpd
from shapely.geometry import Point, Polygon
from emiproc.inventories.utils import crop_with_shape
from emiproc.regrid import (
    geoserie_intersection,
)
from emiproc.inventories import Inventory
from emiproc.tests_utils import WEIGHTS_DIR


xmin = 0
dx = 1
nx = 10
ymin = 0
dy = 0.5
ny = 20

xs = np.arange(
    xmin,
    xmin + nx * dx,
    step=dx,
)
ys = np.arange(
    ymin,
    ymin + ny * dy,
    step=dy,
)


grid = gpd.GeoSeries(
    [
        Polygon(
            (
                (x, y),
                (x, y + dy),
                (x + dx, y + dy),
                (x + dx, y),
            )
        )
        for y in reversed(ys)
        for x in xs
    ]
)

poly = Polygon(((5.5, 7.5), (5.5, 3.5), (2.5, 3.5), (2.5, 8.5)))
non_convex_poly = Polygon(
    (
        (5.5, 7.5),
        (5.5, 3.5),
        (2.5, 3.5),
        (2.5, 6.5),
        (4.5, 6.5),
        (4.5, 7.5),
        (2.5, 7.5),
        (2.5, 8.5),
    )
)

inv = Inventory.from_gdf(
    gpd.GeoDataFrame({("main", "val"): np.linspace(0, 1, len(grid))}, geometry=grid)
)

point_inside = Point((4.5, 5.5))
point_outside = Point((1.5, 5.5))
inv_with_point_sources = Inventory.from_gdf(
    gpd.GeoDataFrame({("main", "val"): np.linspace(0, 1, len(grid))}, geometry=grid),
    gdfs={
        "pnt_sources": gpd.GeoDataFrame(
            {"val": [1, 2]}, geometry=[point_inside, point_outside]
        )
    },
)


def test_basic_crop():
    intersected_shapes, weigths = geoserie_intersection(
        grid, poly, keep_outside=False, drop_unused=False
    )


def test_with_modify_grid():
    intersected_shapes, weigths = geoserie_intersection(
        grid, poly, keep_outside=False, drop_unused=False
    )


def test_crop_inventory():
    inv_cropped = crop_with_shape(inv, poly, keep_outside=False)


def test_crop_inventory_outside():
    inv_cropped = crop_with_shape(inv, poly, keep_outside=True)


def test_crop_inventory_inside_pnt_sources():
    inv_cropped = crop_with_shape(inv_with_point_sources, poly, keep_outside=True)
    assert len(inv_cropped.gdfs["pnt_sources"]) == 1
    assert inv_cropped.gdfs["pnt_sources"]["val"].iloc[0] == 2


def test_crop_inventory_outside_pnt_sources():
    inv_cropped = crop_with_shape(inv_with_point_sources, poly, keep_outside=False)
    assert len(inv_cropped.gdfs["pnt_sources"]) == 1
    assert inv_cropped.gdfs["pnt_sources"]["val"].iloc[0] == 1


def test_read_weights():
    w_file = WEIGHTS_DIR / ".emiproc__test_read_weights"
    if w_file.with_suffix(".gdb").is_dir():
        shutil.rmtree(w_file.with_suffix(".gdb"))
    w_file.with_suffix(".npy").unlink(missing_ok=True)

    inv_cropped_1 = crop_with_shape(inv, poly, weight_file=w_file)
    # This time the weights should have been created
    inv_cropped_2 = crop_with_shape(inv, poly, weight_file=w_file)
    assert all(inv_cropped_1.gdf.index == inv_cropped_2.gdf.index)
    # Check that the polygons are the same
    assert inv_cropped_1.gdf.geometry.equals(inv_cropped_2.gdf.geometry)
    # Check that there is not none geomtery
    assert all([t is not None for t in inv_cropped_1.gdf.geometry])


def test_read_weights_modified_grid():
    w_file = WEIGHTS_DIR / ".emiproc_test_read_weights_modified_grid"
    if w_file.with_suffix(".gdb").is_dir():
        shutil.rmtree(w_file.with_suffix(".gdb"))
    w_file.with_suffix(".npy").unlink(missing_ok=True)
    inv_cropped_1 = crop_with_shape(inv, poly, weight_file=w_file, modify_grid=True)
    # This time the weights should have been created
    inv_cropped_2 = crop_with_shape(inv, poly, weight_file=w_file, modify_grid=True)

    assert all(inv_cropped_1.gdf.index == inv_cropped_2.gdf.index)
    # Check that the polygons are the same
    assert inv_cropped_1.gdf.geometry.equals(inv_cropped_2.gdf.geometry)
    # Check that there is not none geomtery
    assert all([t is not None for t in inv_cropped_1.gdf.geometry])


def test_non_convex_poly():
    intersected_shapes, weigths = geoserie_intersection(
        grid, non_convex_poly, keep_outside=True, drop_unused=True
    )

    cropped_inv = crop_with_shape(
        inv_with_point_sources, non_convex_poly, keep_outside=True, modify_grid=True
    )

    cropped_inv.gdf.geom_type



"""Utilities for the diffenrent inventory."""

from __future__ import annotations
import collections
import itertools
import json
import logging
from os import PathLike
from pathlib import Path
from typing import TYPE_CHECKING

import fiona
import geopandas as gpd
import pandas as pd
import numpy as np
import xarray as xr

from shapely.geometry import Point, MultiPolygon, Polygon
from emiproc.grids import Grid

from emiproc.regrid import geoserie_intersection
from emiproc.profiles.operators import (
    add_profiles,
    get_weights_of_gdf_profiles,
    group_profiles_indexes,
)

if TYPE_CHECKING:
    from emiproc.inventories import Inventory, Category, Substance


def list_categories(file: PathLike) -> list[str]:
    """Return the emission categories for the desired dataset."""
    return fiona.listlayers(file)


def load_category(file: PathLike, category: str) -> gpd.GeoDataFrame:
    """Load the geodataframe of the requested category."""
    return gpd.read_file(
        file,
        layer=category,
    )


def process_emission_category(
    file: PathLike,
    category: str,
    convert_lines_to_polygons: bool = True,
    line_width: float = 10,
) -> gpd.GeoDataFrame:
    """Process an emission category.

    Is used for Zurich Mapluft but could be adapted to other inventories.

    Does the following:

    * Load the categorie from the file
    * Convert the line shapes to polygon using :py:arg:`line_width` .

    The absolute Emission for each shape are in unit: [kg / a].
    We convert here the line emissions in polygons.

    :return: A list of tuples containing the shapes and their emission value.
    """
    gdf = load_category(file, category)

    # Sometimes it is written in big and sometimes in small 🤷
    if convert_lines_to_polygons and ("Shape_Length" in gdf or "SHAPE_Length" in gdf):
        # Convert lines into Polygon

        vector_geometry = gdf.geometry.buffer(line_width, cap_style=3)

        gdf.geometry = vector_geometry

    return gdf


def validate_group(categories_groups: dict[str, list[str]], all_categories: list[str]):
    """Check the validity of a group.

    The categories_groups is a mapping from group_name: list_of_categories_in_group.
    all_categories is a list contaning all categories to check.

    This will check that :

    * A category is not included in two groups.
    * That all categories are inside a group.

    :arg categories_groups: The mapping of how the categories are to be groupped.
    :arg all_categories: A list of the name of the categories in the inventar.
    :raise  ValueError: if the mapping is not valid.
    """

    all_categories_in_groupes = itertools.chain(*categories_groups.values())

    c_groups = collections.Counter(all_categories_in_groupes)
    c_all = collections.Counter(all_categories)
    if c_groups != c_all:
        raise ValueError(
            "Categories in 'categories_groups' are not matching 'all_categories'."
            " Problem cause by "
            f"duplicates or not in data: {c_groups - c_all} or "
            f"missing in group mapping: {c_all - c_groups}"
        )


def crop_with_shape(
    inv: Inventory,
    shape: Polygon,
    keep_outside: bool = False,
    weight_file: PathLike | None = None,
    modify_grid: bool = False,
) -> Inventory:
    """Crop the inventory with the provided shape.

    Keeps only the part of the emissions that is included inside the shape.
    Emissions at the boundary with the shapes will see their values based on
    the fraction which is inside the shape.

    Point sources located exactly at the boundary
    will have their emissions value divided by 2.

    This might removes categories and substances from the inventory, if they
    are not present anymore !

    :arg inv: The inventory to crop.
    :arg shape: The shape around which to crop the inv.
    :arg keep_outside: Whether to keep only the emissions outside of the shape.
    :arg weight_file: A file in which to store the weights.
        If modify_grid is True, this will also save the shapes of the output.
        However saving/reading those shape can be slower than computing
        them.
    :arg modify_grid: Whether the main grid (the gdf) should be modified.
        Grid cells cropped will disappear.
        Grid cells intersected will be replaced by the intersection with
        the shape.

    .. warning::
        Make sure your shape is in the same crs as the inventory.
    """
    inv_out = inv.copy(no_gdfs=True)

    if weight_file is not None:
        weight_file = Path(weight_file).with_suffix(".npy")

        if modify_grid:
            shapes_file = Path(weight_file).with_suffix(".gdb")

    if inv.gdf is not None:
        # Check if the weights are already computed
        if weight_file is not None and weight_file.is_file():
            weights = np.load(weight_file)
            if modify_grid:
                if not shapes_file.is_dir():
                    raise RuntimeError(
                        f"A {weight_file=} was found but no {shapes_file=}."
                        " Delete the weight file to recompute it."
                    )
                # Load cached shapes
                gdf_cached_shapes: gpd.GeoDataFrame = gpd.read_file(
                    shapes_file, engine="pyogrio"
                )
                # Index was set with cache
                intersection_shapes = gdf_cached_shapes.set_index(
                    "index", drop=True
                ).geometry
        else:
            # Find the weight of the intersection, keep the same geometry
            intersection_shapes, weights = geoserie_intersection(
                inv.geometry, shape, keep_outside=keep_outside, drop_unused=modify_grid
            )
            if weight_file is not None:
                # Save the weight file
                np.save(weight_file, weights)
                if modify_grid:
                    gpd.GeoDataFrame(
                        {"index": intersection_shapes.index},
                        geometry=intersection_shapes,
                        index=intersection_shapes.index,
                    ).to_file(shapes_file, engine="pyogrio")

        inv_out.gdf = gpd.GeoDataFrame(
            {
                col: (
                    # Select the correct values from the shapes
                    inv.gdf.loc[intersection_shapes.index, col].to_numpy()
                    if modify_grid
                    else inv.gdf[col]
                )
                * weights
                for col in inv._gdf_columns
            },
            geometry=(
                intersection_shapes.reset_index(drop=True)
                if modify_grid
                else inv.geometry
            ),
            crs=inv.gdf.crs,
        )
    else:
        inv_out.gdf = None

    inv_out.gdfs = {}
    for cat, gdf in inv.gdfs.items():
        cols = [col for col in inv.substances if col in gdf]
        if not cols:
            # No substance of the inventory is in this category
            # No need to crop anything (cropping this will create accessing error bug later in the loop)
            continue

        mask_points = gdf.geometry.apply(lambda x: isinstance(x, Point))
        if any(mask_points):
            point_geometries = gdf.geometry.loc[mask_points]
            points_gdf = gdf.loc[mask_points].copy()
            # Simply remove the point sources outside
            mask_intersects = point_geometries.intersects(shape)
            mask_boundary = point_geometries.intersects(shape.boundary)

            # Takes shapes of interest (inside/outside and the boundary)
            mask_shapes = (
                ~mask_intersects if keep_outside else mask_intersects
            ) | mask_boundary

            # Points at the boundary are divided by 2
            points_gdf.loc[mask_boundary, cols] /= 2.0
            inv_out.add_gdf(cat, points_gdf.loc[mask_shapes].reset_index(drop=True))
        if not all(mask_points):
            # We keep crop the geometry
            polys_gdf = gdf.loc[~mask_points]
            new_geometry, weights = geoserie_intersection(
                polys_gdf.geometry, shape, keep_outside=keep_outside, drop_unused=False
            )
            mask_non_zero = weights > 0
            inv_out.add_gdf(
                cat,
                gpd.GeoDataFrame(
                    {
                        col: polys_gdf.loc[mask_non_zero, col] * weights[mask_non_zero]
                        for col in cols
                    },
                    geometry=new_geometry[mask_non_zero],
                    crs=gdf.crs,
                ).reset_index(drop=True),
            )

    inv_out.history.append(f"Cropped using {shape=}, {keep_outside=}")
    return inv_out


def group_categories(
    inv: Inventory,
    categories_group: dict[str, list[str]],
    ignore_missing: bool = False,
) -> Inventory:
    """Group the categories of an inventory in new categories.

    Total emissions are summed among the categories.
    Vertical profiles are weightly averaged, such that the profiles
    of a category with higher emission is more taken into account.
    Point sources vertical profiles are not modified.

    :arg inv: The Inventory to group.
    :arg categories_group: A mapping of which groups should be greated
        out of which categries. This will be checked using
        :py:func:`validate_group` .
    :arg ignore_missing: If True, function will work even if some categories
        from the mapping are not in the inventory.
        Ex. ``{"group1": ["cat1", "cat2"], "group2": ["cat3", "cat4"]}``
        If ``cat3`` is not in the inventory, the function will work as if
        ``{"group1": ["cat1", "cat2"], "group2": ["cat4"]}`` was passed.
    """
    if ignore_missing:
        # Remove the missing categories
        categories_group = {
            group: [cat for cat in categories if cat in inv.categories]
            for group, categories in categories_group.items()
        }

    validate_group(categories_group, inv.categories)

    out_inv = inv.copy(no_gdfs=True)

    if inv.gdf is not None:
        out_inv.gdf = gpd.GeoDataFrame(
            {
                # Sum all the categories containing that substance
                (group, substance): group_sum
                for substance in inv.substances
                for group, categories in categories_group.items()
                # Only add the group if there are some non zero value
                if np.any(
                    group_sum := sum(
                        (
                            inv.gdf[(cat, substance)]
                            for cat in categories
                            if (cat, substance) in inv.gdf
                        )
                    )
                )
            },
            geometry=inv.gdf.geometry,
            crs=inv.crs,
        )
    else:
        out_inv.gdf = None
    # Add the additional gdfs as well
    # Merging the categories directly
    out_inv.gdfs = {}
    for group, categories in categories_group.items():
        group_gdfs = [inv.gdfs[cat] for cat in categories if cat in inv.gdfs]

        # Add missing profile -1 to the gdfs having no profiles column
        for profile_col in ["__v_profile__", "__t_profile__"]:
            if any((profile_col in gdf.columns for gdf in group_gdfs)):
                for gdf in group_gdfs:
                    if profile_col not in gdf.columns:
                        gdf[profile_col] = -1

        if group_gdfs:
            if len(group_gdfs) == 1:
                # Case only one dataframe is registered
                out_inv.gdfs[group] = group_gdfs[0]
            else:
                # Otherwise merge the dataframes
                df_merged = pd.concat(group_gdfs, ignore_index=True)
                # Na values are no emission, replaces nan by 0
                out_inv.gdfs[group] = df_merged.mask(pd.isna(df_merged), 0.0)

    # Group the vertical profiles
    # we group only on the gdf, as the gdfs will keep their own profiles
    for profiles_name, profiles_indexes_name in [
        ("v_profiles", "v_profiles_indexes"),
        ("t_profiles_groups", "t_profiles_indexes"),
    ]:
        profiles = getattr(inv, profiles_name)
        profiles_indexes: xr.DataArray = getattr(inv, profiles_indexes_name)

        if (
            profiles is not None
            # if they don't depend on category, we don't need to create new profiles
            and "category" in profiles_indexes.dims
        ):
            new_profiles, new_indices = group_profiles_indexes(
                profiles,
                profiles_indexes,
                indexes_weights=get_weights_of_gdf_profiles(inv.gdf, profiles_indexes),
                categories_group=categories_group,
                groupping_dimension="category",
            )

            out_inv.set_profiles(new_profiles, new_indices)
            out_inv.history.append(
                f"Generated new {profiles_indexes_name} from groupping."
            )

    out_inv.history.append(f"groupped from {inv.categories} to {out_inv.categories}")

    return out_inv


def group_substances(
    inv: Inventory,
    substances_group: dict[str, list[str]],
    ignore_missing: bool = False,
) -> Inventory:
    """Group the substances of an inventory in new substances.

    Simalar to :py:func:`group_categories` but for substances.

    """
    if ignore_missing:
        # Remove the missing categories
        substances_group = {
            group: [sub for sub in substances if sub in inv.substances]
            for group, substances in substances_group.items()
        }

    validate_group(substances_group, inv.substances)

    out_inv = inv.copy(no_gdfs=True)

    if inv.gdf is not None:
        out_inv.gdf = gpd.GeoDataFrame(
            {
                # Sum all the substances containing that category
                (cat, group): group_sum
                for cat in inv.categories
                for group, substances in substances_group.items()
                # Only add the group if there are some non zero value
                if np.any(
                    group_sum := sum(
                        (
                            inv.gdf[(cat, sub)]
                            for sub in substances
                            if (cat, sub) in inv.gdf
                        )
                    )
                )
            },
            geometry=inv.gdf.geometry,
            crs=inv.crs,
        )
    else:
        out_inv.gdf = None
    # Add the additional gdfs as well
    # Merging the categories directly
    out_inv.gdfs = {}
    for cat, gdf in inv.gdfs.items():
        for group, substances in substances_group.items():
            columns_to_group = [sub for sub in substances if sub in gdf.columns]
            new_gdf = gdf.copy(deep=True)
            if columns_to_group:
                new_gdf[group] = new_gdf[columns_to_group].sum(axis=1)
                new_gdf.drop(columns=columns_to_group, inplace=True)
            out_inv.gdfs[cat] = new_gdf

    # Group the vertical profiles
    # we group only on the gdf, as the gdfs will keep their own profiles
    for profiles_name, profiles_indexes_name in [
        ("v_profiles", "v_profiles_indexes"),
        ("t_profiles_groups", "t_profiles_indexes"),
    ]:
        profiles = getattr(inv, profiles_name)
        profiles_indexes: xr.DataArray = getattr(inv, profiles_indexes_name)

        out_profiles = getattr(out_inv, profiles_name)

        if (
            profiles is not None
            # if they don't depend on category, we don't need to create new profiles
            and "substance" in profiles_indexes.dims
        ):
            new_profiles, new_indices = group_profiles_indexes(
                profiles,
                profiles_indexes,
                indexes_weights=get_weights_of_gdf_profiles(inv.gdf, profiles_indexes),
                categories_group=substances_group,
                groupping_dimension="substance",
            )

            # Offset the indexes for merging with the profiles
            new_indices = xr.where(
                new_indices != -1,
                new_indices + len(profiles),
                -1,
            )
            out_profiles += new_profiles
            # Replace the old indexes by the new
            setattr(out_inv, profiles_name, out_profiles)
            setattr(out_inv, profiles_indexes_name, new_indices)
            out_inv.history.append(
                f"Generated new {profiles_indexes_name} from groupping."
            )

    out_inv.history.append(f"groupped from {inv.categories} to {out_inv.categories}")

    return out_inv


def add_inventories(inv: Inventory, other_inv: Inventory) -> Inventory:
    """Add inventories together.

    The following conditions must be required:

    * if the two invs have a gdf, they must be on the same grid

    :arg inv: The first inventory.
    :arg other_inv: The second inventory.
    """
    logger = logging.getLogger("emiproc.add_inventories")

    if inv.gdf is None and other_inv.gdf is not None:
        # as we want to put everything on inv.gdf later for simplicity
        return add_inventories(other_inv, inv)

    # Check that the two inventories are on the same grid
    if (
        inv.gdf is not None
        and other_inv.gdf is not None
        and not np.all(
            gpd.GeoSeries.geom_equals(inv.gdf.geometry, other_inv.gdf.geometry)
        )
    ):
        raise ValueError("Grids of the two inventories are not the same.")

    if inv.crs != other_inv.crs:
        raise ValueError("CRS of both inventories differ.")

    out_inv = inv.copy(no_gdfs=True)

    if inv.gdf is not None and other_inv.gdf is not None:
        # Check that they somehow have the same grid
        if not np.all(
            gpd.GeoSeries.geom_equals(inv.gdf.geometry, other_inv.gdf.geometry)
        ):
            raise ValueError(
                "Grids of the gdf of the two inventories are not the same."
            )
        # Get the columns for the new gdf
        cols_a = inv._gdf_columns
        cols_b = other_inv._gdf_columns
        all_cols = set(cols_a) | set(cols_b)

        # Sum the two gdfs
        gdf = gpd.GeoDataFrame(
            {
                col: (inv.gdf[col] if col in inv.gdf else 0)
                + (other_inv.gdf[col] if col in other_inv.gdf else 0)
                for col in all_cols
            },
            geometry=inv.gdf.geometry,
            crs=inv.crs,
        )
    elif inv.gdf is not None:
        # Just copy the gdf
        gdf = inv.gdf.copy()
    else:
        gdf = None

    out_inv.gdf = gdf

    # Process the gdfs
    gdfs = {}
    for cat, gdf in itertools.chain(inv.gdfs.items(), other_inv.gdfs.items()):
        if cat in gdfs:
            df_merged = pd.concat([gdfs[cat], gdf], ignore_index=True)
            # Na values are no emission, replaces nan by 0
            gdfs[cat] = df_merged.mask(pd.isna(df_merged), 0.0)
        else:
            gdfs[cat] = gdf.copy()
    out_inv.gdfs = gdfs

    for profile_name, indexes_name in [
        ("v_profiles", "v_profiles_indexes"),
        ("t_profiles_groups", "t_profiles_indexes"),
    ]:
        profiles = getattr(inv, profile_name, None)
        indexes = getattr(inv, indexes_name, None)
        other_profiles = getattr(other_inv, profile_name, None)
        other_indexes = getattr(other_inv, indexes_name, None)

        if profiles is not None and other_profiles is not None:
            if profile_name == "t_profiles_groups":
                new_profiles, new_indexes = add_profiles(inv, other_inv)
            else:
                raise NotImplementedError(
                    f"Adding {profile_name} is not implemented yes."
                )
        elif profiles is not None:
            new_profiles, new_indexes = profiles, indexes
        elif other_profiles is not None:
            new_profiles, new_indexes = other_profiles, other_indexes
        else:
            # No profiles
            continue
        out_inv.set_profiles(new_profiles, new_indexes)

    out_inv.history.append(f"Added inventory {other_inv}")

    return out_inv


def get_total_emissions(inv: Inventory) -> dict[str, dict[str, float]]:
    """Get the total emissions from the inventory.

    :arg inv: The inventory from which to get the total emissions.

    :return: A dictionary mapping substances to another dictionary
        which maps categories to values.
        A '__total__' key will be created in each substnace mapping
        with the total of all the categories.

        For exemple ::

            {
                "CO2": {
                    "cat1": 3.2,
                    "cat2": 4.3,
                    "__total__": 7.5,
                },
                "CH4": {
                    "cat2": 2.1,
                    "__total__": 2.1,
                },
                ...
            }
    """

    # Preapre the output dictionary
    out_dic = {sub: {} for sub in inv.substances}

    # First look for the emissions in the gdf
    for cat, sub in inv._gdf_columns:
        # Calculate the total sum
        out_dic[sub][cat] = inv.gdf[(cat, sub)].sum()

    # Second look for the emissions in the gdfs
    for cat, gdf in inv.gdfs.items():
        for sub in inv.substances:
            if sub not in gdf.columns:
                # this category does not have the substance
                continue
            if cat not in out_dic[sub]:
                out_dic[sub][cat] = 0
            # Add the total emissions
            out_dic[sub][cat] += gdf[sub].sum()
    # Add the total
    for dic in out_dic.values():
        dic["__total__"] = sum(dic.values())

    return out_dic


def scale_inventory(
    inv: Inventory, scaling_dict: dict[str, dict[str, float]] | float
) -> Inventory:
    """Get the total emissions from the inventory.

    :arg inv: The inventory from which to get the total emissions.

    :arg scaling_dict: A dictionary mapping substances to another dictionary
        which maps categories to values.
        The values must be scaling factor, which will multiply all the objects
        from the inventory with matching categories and substance.

        If a category/substance is not in the dict, it will not be scaled.
        For exemple ::

            {
                "CO2": {
                    "cat1": 1.3,
                    "cat2": 0.7,
                },
                "CH4": {
                    "cat2": 1.2,
                },
            }

        Alternatively, a float can be given, which will scale all the emissions.

        If you have a gridded inventory (no gdfs emissions), you can also scale
        each grid cell individually by giving arrays of the same length as the
        number of grid cells.

    :return: A new inventory with its emission values rescaled.
    """
    # Deep copy of the inventory
    inv = inv.copy()

    # Create the scaling dict if a float was given
    if isinstance(scaling_dict, int):
        scaling_dict = float(scaling_dict)
    if isinstance(scaling_dict, float):
        scaling_dict = {
            sub: {cat: scaling_dict for cat in inv.categories} for sub in inv.substances
        }

    # Iterate over the scaling dict to multiply the values
    for sub, sub_dict in scaling_dict.items():
        for cat, scaling_factor in sub_dict.items():
            if inv.gdf is not None and (cat, sub) in inv.gdf.columns:
                inv.gdf[(cat, sub)] *= scaling_factor
            if inv.gdfs is not None and cat in inv.gdfs.keys() and sub in inv.gdfs[cat]:
                inv.gdfs[cat][sub] *= scaling_factor

    inv.history.append(f"Rescaled using {scaling_dict=}")
    return inv


def combine_inventories(
    inv_inside: Inventory,
    inv_outside: Inventory,
    separated_shape: Polygon,
    output_grid: Grid | None = None,
):
    """Combine two inventories and use a shape as the boundary between the two inventories.

    .. note::
        This is not implemented yet.
        One should be careful about this implementation steps.
        Some comments are already in the code for what should be done.
    """
    # Crop the inventories around the shape

    # Drop substances/categories if only in one of them
    # Or assume they are 0

    # Check if an output grid was given
    # If yes, remap both on new grid
    # if false remap one of the two on the other: Decide how

    # Now that the two are on the same grid, we can simply

    raise NotImplementedError()


def drop(
    inv: Inventory,
    substances: list[Substance] = [],
    categories: list[Category] = [],
    keep_instead_of_drop: bool = False,
) -> Inventory:
    """Drop substances and categories from an inventory.

    This will drop all the emissions related to the substances and categories
    from the inventory.

    If both substances and categories are specified, the emissions will be dropped
    if they are in any of the two lists.

    :arg inv: The inventory to drop from.
    :arg substances: The substances to drop.
    :arg categories: The categories to drop.
    :arg keep_instead_of_drop: If True, the substances and categories will be kept
        instead of being dropped.
    """

    # Check the types
    for var_name, var in [("substances", substances), ("categories", categories)]:
        if not isinstance(var, list):
            raise TypeError(f"{var_name=} should be a list.")
        if not all(isinstance(sub, str) for sub in var):
            raise TypeError(f"{var_name=} should be a list of strings.")

    if keep_instead_of_drop:
        # Keep instead of drop
        if substances:
            substances = [sub for sub in inv.substances if sub not in substances]
        if categories:
            categories = [cat for cat in inv.categories if cat not in categories]

    # Deep copy of the inventory
    out_inv = inv.copy(no_gdfs=True)

    gdf_cols_to_drop = [
        (cat, sub)
        for cat, sub in inv._gdf_columns
        if cat in categories or sub in substances
    ]

    out_inv.gdf = (
        inv.gdf.drop(columns=gdf_cols_to_drop) if inv.gdf is not None else None
    )

    # Process the gdfs
    out_inv.gdfs = {}
    for cat, gdf in inv.gdfs.items():
        if cat in categories:
            continue
        out_inv.gdfs[cat] = gdf.drop(
            columns=[sub for sub in substances if sub in gdf.columns]
        )

    out_inv.history.append(f"Dropped {substances=} and {categories=}")
    return out_inv



from __future__ import annotations
from datetime import datetime, timedelta
from enum import Enum, auto
import logging
from os import PathLike
from pathlib import Path
from zoneinfo import ZoneInfo
import pandas as pd
import xarray as xr
import numpy as np
from emiproc.exports.netcdf import DEFAULT_NC_ATTRIBUTES
from emiproc.grids import ICONGrid
from emiproc.inventories import Inventory
from emiproc.profiles.temporal_profiles import (
    DailyProfile,
    MounthsProfile,
    TemporalProfile,
    WeeklyProfile,
    HourOfYearProfile,
    HourOfLeapYearProfile,
    create_scaling_factors_time_serie,
)
from emiproc.profiles.vertical_profiles import (
    VerticalProfile,
    VerticalProfiles,
    resample_vertical_profiles,
)
from emiproc.utilities import (
    SEC_PER_HOUR,
    SEC_PER_YR,
    get_timezone_mask,
)
from emiproc.profiles.utils import get_desired_profile_index


class TemporalProfilesTypes(Enum):
    """Possible temporal profiles for OEM.

    :param HOUR_OF_YEAR:  Every hour gets a scaling factor
    :param THREE_CYCLES:  Three cycles (hour of day, day of week, month of year)
    """

    HOUR_OF_YEAR = 0
    THREE_CYCLES = 3


def get_constant_time_profile(
    type: TemporalProfilesTypes = TemporalProfilesTypes.THREE_CYCLES,
    year: int | None = None,
) -> list[TemporalProfile]:
    """Get a constant time profile compatible with ICON-OEM.

    Emits the same at every time.

    Contains three profiles: hour of day, day of week, month of year.

    """
    if type == TemporalProfilesTypes.THREE_CYCLES:
        return [
            DailyProfile(),  # Hour of day
            WeeklyProfile(),  # Day of week
            MounthsProfile(),  # Month of year
        ]
    elif type == TemporalProfilesTypes.HOUR_OF_YEAR:
        if year is None:
            raise ValueError("You must provide a year for the HOUR_OF_YEAR option.")
        # Check if it is a leap year
        if year % 4 == 0 and (year % 100 != 0 or year % 400 == 0):
            return [HourOfLeapYearProfile()]
        else:
            return [HourOfYearProfile()]
    else:
        raise NotImplementedError(f"{type} is not implemented.")


def export_icon_oem(
    inv: Inventory,
    icon_grid_file: PathLike,
    output_dir: PathLike,
    group_dict: dict[str, list[str]] = {},
    temporal_profiles_type: TemporalProfilesTypes = TemporalProfilesTypes.THREE_CYCLES,
    year: int | None = None,
    nc_attributes: dict[str, str] = DEFAULT_NC_ATTRIBUTES,
    substances: list[str] | None = None,
):
    """Export to a netcdf file for ICON OEM.

    The inventory should have already been remapped to the
    :py:class:`emiproc.grids.ICONGrid` .

    For ICON-OEM you will need to add in the ICON namelist the path the
    files produced by this module::


        ! oem_nml: online emission module ---------------------------------------------
        &oemctrl_nml
        gridded_emissions_nc        =   '${OEMDIR}/oem_gridded_emissions.nc'
        vertical_profile_nc         =   '${OEMDIR}/vertical_profiles.nc'
        hour_of_day_nc              =   '${OEMDIR}/hourofday.nc'
        day_of_week_nc              =   '${OEMDIR}/dayofweek.nc'
        month_of_year_nc            =   '${OEMDIR}/monthofyear.nc'
        ! If you use the hour of year profile, use this instead of the three above
        ! hour_of_year_nc             =   '${OEMDIR}/hourofyear.nc'
        /


    Values will be converted from emiproc units: `kg/y` to
    OEM units  `kg/m2/s` .
    The grid cell area given in the icon grid file is used for this conversion,
    and 365.25 days per year.

    Temporal profiles are adapted to the different countries present in the data.
    Shifts for local time are applied to the countries individually.
    Grid cells are assigned to a country using the timezone mask from
    :py:func:`emiproc.utilities.get_timezone_mask` and country ids are
    set as the `country_id` attribute of the output NetCDF file.

    :arg inv: The inventory to export.
    :arg icon_grid_file: The icon grid file.
    :arg output_dir: The output directory.
    :arg group_dict: If you groupped some categories, you can optionally
        add the groupping in the metadata.
    :arg country_resolution: The resolution
        can be either '10m', '50m' or '110m'
    :arg temporal_profiles_type: The type of temporal profiles to use.
        Can be either :py:class:`~emiproc.exports.icon.TemporalProfilesTypes.HOUR_OF_YEAR`
        or :py:class:`~emiproc.exports.icon.TemporalProfilesTypes.THREE_CYCLES`
    :arg year: The year to use for the temporal profiles. This is mandatory
        only with `temporal_profiles_type` set to
        :py:class:`~emiproc.exports.icon.TemporalProfilesTypes.HOUR_OF_YEAR`
    :arg nc_attributes: The attributes to add to the netcdf file.
    :arg substances: The substances to export. If None, all substances of the inv.

    """
    logger = logging.getLogger("emiproc.export_icon_oem")

    icon_grid_file = Path(icon_grid_file)
    output_dir = Path(output_dir)
    output_dir.mkdir(exist_ok=True, parents=True)
    # Load the output xarray

    ds_out: xr.Dataset = xr.load_dataset(icon_grid_file)
    time_profiles: dict[str, list[TemporalProfile]] = {}
    vertical_profiles: dict[str, VerticalProfile] = {}

    # Check that the inventory has the same amount of cells
    # as the icon grid
    if len(inv.gdf) != ds_out["cell"].size:
        raise ValueError(
            f"The inventory has {len(inv.gdf)} cells, but the icon grid has"
            f" {ds_out['cell'].size} cells."
        )

    for categorie, sub in inv._gdf_columns:
        if substances is not None and sub not in substances:
            continue
        name = f"{categorie}-{sub}"

        # Convert from kg/year to kg/m2/s
        emissions = (
            inv.gdf[(categorie, sub)].to_numpy() / ds_out["cell_area"] / SEC_PER_YR
        )

        attributes = {
            "units": "kg/m2/s",
            "standard_name": name,
            "long_name": f"Emission of {sub} from {categorie}",
            "created_by_emiproc": f"{datetime.now()}",
        }
        if group_dict:
            attributes["group_made_from"] = f"{group_dict[categorie]}"

        emission_with_metadata = emissions.assign_attrs(attributes)

        ds_out = ds_out.assign({name: emission_with_metadata})

        if inv.v_profiles is not None:
            profile_index = get_desired_profile_index(
                inv.v_profiles_indexes, cat=categorie, sub=sub, type="gridded"
            )
            vertical_profiles[name] = inv.v_profiles[profile_index]

        if inv.t_profiles_groups is not None:
            profile_index = get_desired_profile_index(
                inv.t_profiles_indexes, cat=categorie, sub=sub, type="gridded"
            )
            time_profiles[name] = inv.t_profiles_groups[profile_index]

    # Find the proper country codes
    mask_file = (output_dir / f".emiproc_tz_mask_{icon_grid_file.stem}").with_suffix(
        ".npy"
    )

    tz_mask = None

    if mask_file.is_file():
        tz_mask = np.load(mask_file)
    else:
        icon_grid = ICONGrid(icon_grid_file)
        tz_mask = get_timezone_mask(icon_grid)
        np.save(mask_file, tz_mask)

    # Get the countries as integers
    unique_timezones, tz_mask_int = np.unique(tz_mask.reshape(-1), return_inverse=True)

    # Save the profiles
    if time_profiles:
        make_icon_time_profiles(
            time_profiles=time_profiles,
            time_zones=unique_timezones,
            profiles_type=temporal_profiles_type,
            year=year,
            out_dir=output_dir,
            nc_attrs=nc_attributes,
        )
    if vertical_profiles:
        make_icon_vertical_profiles(
            vertical_profiles, out_dir=output_dir, nc_attrs=nc_attributes
        )

    # Add the country ids variable for oem
    ds_out = ds_out.assign(
        {
            "country_ids": (
                "cell",
                tz_mask_int,
                {
                    "standard_name": "country_ids",
                    "long_name": "Timezone of the cell",
                    "history": f"Added by emiproc",
                    "country_resolution": f"country_resolution",
                },
            ),
            "timezone_of_country": (
                "timezones",
                unique_timezones,
                {
                    "standard_name": "timezone_of_country",
                    "long_name": "Timezone corresponding the country id",
                    "history": f"Added by emiproc",
                },
            ),
        }
    )
    # Save the emissions
    ds_out.to_netcdf(output_dir / "oem_gridded_emissions.nc")

    logger.info(f"Exported inventory to {output_dir}.")


def make_icon_time_profiles(
    time_profiles: dict[str, list[TemporalProfile]],
    time_zones: list[str],
    profiles_type: TemporalProfilesTypes = TemporalProfilesTypes.THREE_CYCLES,
    year: int | None = None,
    out_dir: PathLike | None = None,
    nc_attrs: dict[str, str] = DEFAULT_NC_ATTRIBUTES,
) -> dict[str, xr.Dataset]:
    """Make the profiles in the format for icon oem.

    :arg time_profiles: A dictionary with
        the names of variables in the file as keys and
        the profiles as values.
    :arg time_zones: A list with valid timezones names that have to be
        included.
    :arg profiles_type: The type of profiles to use.
    :arg year: Used for the HOUR_OF_YEAR option.
    :arg out_dir: The directory where to export the files.
        If None, the files are not saved and dataset returned.


    .. note::
        OEM can differentiate profiles based on the grid cell.
        It tries to group grid cells in what it calls "countries".
        This should not be mixed with the real countries.
        The countries identifiers should match between the emission files.

    .. warning::
        Currently the same profiles are used for all the countries.
        Only the shifts are different.
    """

    date_of_shift = datetime(
        year=year if year is not None else datetime.now().year, month=1, day=1
    )
    shifts = np.array(
        [
            # Calculate the country shifts for the grid cells
            int(ZoneInfo(code).utcoffset(date_of_shift).seconds / SEC_PER_HOUR)
            for code in time_zones
        ]
    )
    max_shift = int(max(np.abs(shifts)))

    profiles_type = TemporalProfilesTypes(profiles_type)

    if profiles_type == TemporalProfilesTypes.THREE_CYCLES:
        nc_attrs["title"] = "Hour of day profiles"
        hourofday = xr.Dataset(attrs=nc_attrs.copy())
        nc_attrs["title"] = "Day of week profiles"
        dayofweek = xr.Dataset(attrs=nc_attrs.copy())
        nc_attrs["title"] = "Month of year profiles"
        monthofyear = xr.Dataset(attrs=nc_attrs.copy())
    else:
        nc_attrs["title"] = "Hour of year profiles"
        hourofyear = xr.Dataset(attrs=nc_attrs)

    var_metadata = lambda var_name, profile_name: {
        "units": "1",
        "long_name": f"{profile_name} scaling factors for {var_name}",
    }

    for key in time_profiles:
        if profiles_type == TemporalProfilesTypes.THREE_CYCLES:
            for profile in time_profiles[key]:
                if not issubclass(type(profile), TemporalProfile):
                    raise TypeError(f"{profile} from {key} is not a TemporalProfile")
                if profile.n_profiles != 1:
                    raise ValueError(
                        f"{profile} from {key} has {profile.n_profiles} profiles."
                        " Only one profile is allowed for the THREE_CYCLES option."
                    )
                scaling_factors = profile.ratios.reshape(-1) * profile.size
                if isinstance(profile, DailyProfile):
                    ds = hourofday
                    # Use the shifts in the intervals
                    data = np.asarray(
                        [np.roll(scaling_factors, -shift) for shift in shifts]
                    )
                    dim = "hourofday"

                else:
                    if isinstance(profile, WeeklyProfile):
                        ds = dayofweek
                        dim = "dayofweek"
                    elif isinstance(profile, MounthsProfile):
                        ds = monthofyear
                        dim = "monthofyear"
                    else:
                        raise TypeError(
                            f"{profile} from {key} is not on of the three profiles:"
                            " DailyProfile, WeeklyProfile, MounthsProfile. You can use"
                            " the HOUR of YEAR option to have scaling with this type"
                            " of profile."
                        )
                    data = np.asarray([scaling_factors for _ in time_zones])

                ds[key] = xr.DataArray(
                    data.T,
                    dims=[dim, "country"],
                    attrs=var_metadata(key, dim),
                )
        elif profiles_type == TemporalProfilesTypes.HOUR_OF_YEAR:
            if year is None:
                raise ValueError("You must provide a year for the HOUR_OF_YEAR option.")

            # Use the shifts in the intervals
            dt_start = datetime(year, 1, 1, hour=0)  # - timedelta(hours=max_shift)
            dt_end = datetime(year, 12, 31, hour=23)  # + timedelta(hours=max_shift + 1)

            concatenated_profiles = np.asarray(
                [
                    # Start around the shift and end
                    create_scaling_factors_time_serie(
                        dt_start,
                        dt_end,
                        time_profiles[key],
                        local_tz=tz,
                        freq="h",
                    ).to_numpy()  # [max_shift + shift : -max_shift + shift - 1]
                    for shift, tz in zip(shifts, time_zones)
                ]
            )

            # Apply the shift for each contry
            hourofyear[key] = xr.DataArray(
                data=concatenated_profiles.T,
                coords={
                    "datetime": (
                        ("hourofyear",),
                        pd.date_range(dt_start, dt_end, freq="h"),
                    ),
                    "timezone_of_country": (("country",), time_zones),
                },
                dims=["hourofyear", "country"],
                attrs=var_metadata(key, "hourofyear"),
            )
        else:
            raise NotImplementedError(f"{profiles_type} is not implemented.")

    if profiles_type == TemporalProfilesTypes.HOUR_OF_YEAR:
        dict_ds = {"hourofyear": hourofyear}
    elif profiles_type == TemporalProfilesTypes.THREE_CYCLES:
        dict_ds = {
            "hourofday": hourofday,
            "dayofweek": dayofweek,
            "monthofyear": monthofyear,
        }
    else:
        raise NotImplementedError(f"{profiles_type} is not implemented.")

    for ds in dict_ds.values():
        ds["country"] = [i for i in range(len(time_zones))]
        ds.assign_coords({"timezone_of_country": (("country",), time_zones)})

    if out_dir is not None:
        # Save the files
        out_dir = Path(out_dir)
        out_dir.mkdir(parents=True, exist_ok=True)
        for name, ds in dict_ds.items():
            ds.to_netcdf(out_dir / f"{name}.nc")

    return dict_ds


def make_icon_vertical_profiles(
    vertical_profiles: dict[str, VerticalProfile],
    out_dir: PathLike | None = None,
    nc_attrs: dict[str, str] = DEFAULT_NC_ATTRIBUTES,
) -> xr.Dataset:
    """Create the vertical profiles in the icon format."""

    # Make sure all teh vertical profiles have the same heights
    # oem does not process different heigths
    resampled_profiles = resample_vertical_profiles(*vertical_profiles.values())

    data_vars = {
        key: (
            "level",
            resampled_profiles[i].ratios,
            {
                "long_name": f"vertical scaling factor for sources of {key} category ",
                "units": "1",
            },
        )
        for i, key in enumerate(vertical_profiles.keys())
    }

    # Add the layers
    data_vars["layer_top"] = ("level", resampled_profiles.height)
    layer_bot = np.roll(resampled_profiles.height, 1)
    layer_bot[0] = 0
    data_vars["layer_bot"] = ("level", layer_bot)
    data_vars["layer_mid"] = ("level", (resampled_profiles.height + layer_bot) / 2.0)

    # Create the dataset
    nc_attrs["title"] = "Vertical profiles"
    ds = xr.Dataset(data_vars=data_vars, attrs=nc_attrs)

    if out_dir is not None:
        # Save the files
        out_dir = Path(out_dir)
        out_dir.mkdir(parents=True, exist_ok=True)
        ds.to_netcdf(out_dir / "vertical_profiles.nc")

    return ds



import pytest
from emiproc.utilities import total_emissions_almost_equal

def test_total_emissions_almost_equal():
    # Create a reference dictionary
    ref_dict = {
        'sub1': {'cat1': 10.0, 'cat2': 20.0},
        'sub2': {'cat1': 30.0, 'cat2': 40.0}
    }

    # Test case 1: Total emissions are exactly equal
    assert total_emissions_almost_equal(ref_dict, ref_dict) == True

    # Test case 2: Total emissions are almost equal within the default tolerance
    total_dict_2 = {
        'sub1': {'cat1': 10.00001, 'cat2': 19.99999},
        'sub2': {'cat1': 29.99999, 'cat2': 40.00001}
    }
    assert total_emissions_almost_equal(ref_dict, total_dict_2) == True

    # Test case 3: Total emissions are not almost equal within the default tolerance
    total_dict_3 = {
        'sub1': {'cat1': 10.0001, 'cat2': 19.99},
        'sub2': {'cat1': 29.9999, 'cat2': 40.01}
    }
    assert (total_emissions_almost_equal(ref_dict, total_dict_3) == False)

    # Test case 4: Total emissions have different subcategories
    total_dict_4 = {
        'sub1': {'cat1': 10.0, 'cat2': 20.0},
        'sub3': {'cat1': 30.0, 'cat2': 40.0}
    }
    pytest.raises(ValueError, total_emissions_almost_equal, ref_dict, total_dict_4)

    # Test case 5: Total emissions have different categories
    total_dict_5 = {
        'sub1': {'cat1': 10.0, 'cat3': 20.0},
        'sub2': {'cat1': 30.0, 'cat2': 40.0}
    }
    pytest.raises(ValueError, total_emissions_almost_equal, ref_dict, total_dict_5)



"""Functions related to the WRF model."""

from __future__ import annotations

import itertools
import os
from datetime import datetime
from os import PathLike
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr
from shapely.creation import polygons

from emiproc.exports.utils import get_temporally_scaled_array
from emiproc.grids import WGS84, Grid
from emiproc.inventories import Inventory


class WRF_Grid(Grid):
    """Grid of the wrf model.

    The grid is a pseudo regular grid, in the sense that the grid is regular
    under a certain projection, but is never given in that projection, but on a
    WG84 projection.

    The grid is constucted from the wrfinput file.

    """

    attributes: dict[any, any]

    def __init__(self, grid_filepath: PathLike):
        """Initialize the grid.

        Parameters
        ----------
        grid_filepath : Pathlike
            The path to the grid file.
        """

        grid_filepath = Path(grid_filepath)
        super().__init__(name=grid_filepath.stem, crs=WGS84)

        ds = xr.open_dataset(grid_filepath, engine="netcdf4")
        self.attributes = ds.attrs

        # This will be necessary to reshape the arrays to a 1D array following the
        # emiproc convention
        reshape = lambda x: x.T.reshape(-1)

        # Access the grid coordinates
        center_lon = reshape(ds["XLONG"].isel(Time=0).values)
        center_lat = reshape(ds["XLAT"].isel(Time=0).values)

        self.nx = ds.sizes["west_east"]
        self.ny = ds.sizes["south_north"]

        # Grid vertices are given not at the vertices but at edges
        # It is the place where the wind beteween two cells is calculated (thus U and V)
        lon_u = ds["XLONG_U"].isel(Time=0).values
        lon_v = ds["XLONG_V"].isel(Time=0).values
        left_lon_u = reshape(lon_u[:, :-1])
        right_lon_u = reshape(lon_u[:, 1:])
        bottom_lon_v = reshape(lon_v[:-1, :])
        top_lon_v = reshape(lon_v[1:, :])

        lat_u = ds["XLAT_U"].isel(Time=0).values
        lat_v = ds["XLAT_V"].isel(Time=0).values
        bottom_lat_v = reshape(lat_v[:-1, :])
        top_lat_v = reshape(lat_v[1:, :])
        left_lat_u = reshape(lat_u[:, :-1])
        right_lat_u = reshape(lat_u[:, 1:])

        # Calculate the offsets, to be able to reconstruct the grid vertices
        d_lon_right = right_lon_u - center_lon
        d_lon_left = left_lon_u - center_lon
        d_lon_top = top_lon_v - center_lon
        d_lon_bottom = bottom_lon_v - center_lon

        d_lat_right = right_lat_u - center_lat
        d_lat_left = left_lat_u - center_lat
        d_lat_top = top_lat_v - center_lat
        d_lat_bottom = bottom_lat_v - center_lat

        # Reconstruct the grid vertices
        coords = np.array(
            [
                # Bottom left
                [
                    center_lon + d_lon_left + d_lon_bottom,
                    center_lat + d_lat_left + d_lat_bottom,
                ],
                # Bottom right
                [
                    center_lon + d_lon_right + d_lon_bottom,
                    center_lat + d_lat_right + d_lat_bottom,
                ],
                # Top right
                [
                    center_lon + d_lon_right + d_lon_top,
                    center_lat + d_lat_right + d_lat_top,
                ],
                # Top left
                [
                    center_lon + d_lon_left + d_lon_top,
                    center_lat + d_lat_left + d_lat_top,
                ],
            ]
        )

        coords = np.rollaxis(coords, -1, 0)

        # Create the polygons
        polys = polygons(coords)

        self.cells_as_polylist = polys


def export_wrf_hourly_emissions(
    inv: Inventory,
    grid: WRF_Grid,
    time_range: tuple[datetime | str, datetime | str],
    output_dir: PathLike,
    variable_name: str = "E_{substance}_{category}",
) -> Path:
    """Export the inventory to WRF chemi files.

    .. note::
        When running this function on Windows, the files will be saved with the
        `:` replaced by `-` in the file name, because Windows does not allow `:` in
        file names.

    :param inv: the inventory to export
    :param grid: the grid of the WRF model. The inventory must be on this grid.
    :param time_range: the time range to export the inventory.
    :param output_dir: the directory where to save the files.
    :param variable_name: the name of the variable in the netcdf file.
        You can use the following placeholders:
        - {substance}
        - {category}
        example: "E_{substance}_{category}"

    :return: the directory where the files are saved.

    """

    output_dir = Path(output_dir)
    output_dir.mkdir(parents=True, exist_ok=True)

    # Ensure the inventory is on the same grid as the WRF grid
    assert inv.grid == grid, "The inventory and the grid are not the same"

    # Create the time axis
    time_range = pd.date_range(time_range[0], time_range[1], freq="h")

    da = get_temporally_scaled_array(inv, time_range, sum_over_cells=False)

    # Unstack the datarray to get on the regular 2D grid
    shape = grid.shape
    x_index = np.arange(shape[0])
    y_index = np.arange(shape[1])
    da = da.assign_coords(
        x=("cell", np.repeat(x_index, shape[1])),
        y=("cell", np.tile(y_index, shape[0])),
    )
    da = da.assign_coords(
        cell=pd.MultiIndex.from_arrays([da.x.values, da.y.values], names=["x", "y"])
    )
    da = da.unstack("cell")
    # Rename the dimensions to match the WRF grid
    da = da.rename({"x": "west_east", "y": "south_north"})

    for dt in time_range:
        variables = []
        for cat, sub in itertools.product(inv.categories, inv.substances):
            this_da = (
                da.sel(category=cat, substance=sub, time=dt)
                # Name the variable
                .rename(variable_name.format(substance=sub, category=cat))
                .drop_vars(["substance", "category"])
                # Add the time dimension as a one element dimension
                .expand_dims("Time")
                .expand_dims("emissions_zdim")
            )

            variables.append(this_da)

        ds_at_hour = (
            xr.merge(variables)
            # Transpose to have the dims in the right order
            .transpose("Time", "emissions_zdim", "south_north", "west_east")
            # Progagate the default attributes
            .assign_attrs(grid.attributes)
            # Add emiproc specific attributes
            .assign_attrs(
                {
                    "emiproc": f"This file was created by emiproc on {datetime.now()}",
                    "emiproc_history": f"Created from the inventory {inv.name} with {inv.history=}",
                }
            )
        )

        # Save the dataset
        str_format = "%Y-%m-%d_%H:%M:%S"

        ds_at_hour["Times"] = ("Time", [dt.strftime(str_format).encode()])

        # If windows, we cannot have : in the file name
        if os.name == "nt":
            str_format = "%Y-%m-%d_%H-%M-%S"
        file_name = output_dir / f"wrfchemi_d01_{dt.strftime(str_format)}"
        ds_at_hour.to_netcdf(file_name)

    return output_dir



from emiproc.tests_utils import test_inventories, test_grids, temporal_profiles
import pandas as pd
from emiproc.regrid import calculate_weights_mapping
from emiproc.profiles.operators import get_weights_of_gdf_profiles, remap_profiles
from emiproc.profiles.temporal_profiles import CompositeTemporalProfiles
import xarray as xr
import numpy as np


def test_remap_profiles():

    # Get the test variables required
    inv = test_inventories.inv.copy()
    weights_mapping = calculate_weights_mapping(
        inv.grid.gdf, test_grids.basic_grid_2.gdf
    )
    profiles_indexes = temporal_profiles.indexes_inv_catsubcell
    da_emission_weights = get_weights_of_gdf_profiles(
        inv.gdf,
        profiles_indexes,
    )
    profiles = temporal_profiles.three_composite_profiles

    new_profiles, new_indexes = remap_profiles(
        profiles,
        profiles_indexes,
        da_emission_weights,
        weights_mapping,
    )

    # check that the results are correct
    sel_dict = dict(substance="CH4", category="adf")
    this_new_indices = new_indexes.sel(**sel_dict)
    this_old_indices = profiles_indexes.sel(**sel_dict)

    # All cells have the same index in the given data, make sure still true for the test
    assert np.all(this_old_indices.values == this_old_indices.sel(cell=0).values)
    old_ratios = profiles.ratios[this_old_indices.sel(cell=0).values]
    for new_index in this_new_indices.values:
        assert new_index != -1, "all cells should have a valid index"
        this_ratios = new_profiles.ratios[new_index]
        np.testing.assert_almost_equal(
            this_ratios,
            old_ratios,
            err_msg=f"The new profiles should be the same as the old ones {this_ratios=} != {old_ratios=}",
        )

    # Check that the profiles with no emissions are set to -1
    assert all(
        new_indexes.sel(substance="CH4", category="liku") == -1
    ), "no emissions should give invalid profiles"

    # Check that this profile is a mixture with the correct weights
    sel_dict = dict(substance="CO2", category="liku")
    index_out = new_indexes.sel(**sel_dict, cell=0).values
    assert index_out != -1, "The profile should be valid"
    # This one has just one input cell with a profile
    index_in = profiles_indexes.sel(**sel_dict, cell=4).values
    # Check that the profiles are the same
    np.testing.assert_almost_equal(
        new_profiles.ratios[index_out],
        profiles.ratios[index_in],
        err_msg="The profiles should be the same",
    )

    # Now test some merged data with real combination
    sel_dict = dict(substance="CO2", category="adf")
    index_out = new_indexes.sel(**sel_dict, cell=0).values
    assert index_out != -1, "The profile should be valid"
    # This one has just one input cell with a profile
    index_in = profiles_indexes.sel(**sel_dict, cell=3).values
    assert index_in != -1, "The profile should be valid"
    # Check that the profiles are the same
    np.testing.assert_almost_equal(
        new_profiles.ratios[index_out],
        profiles.ratios[index_in],
        err_msg="The profiles should be the same",
    )
    # This one should be a mixture of the two profiles
    index_out2 = new_indexes.sel(**sel_dict, cell=1).values
    assert index_out2 != -1, "The profile should be valid"

    index_in2 = profiles_indexes.sel(**sel_dict, cell=4).values
    assert index_in2 != -1, "The profile should be valid"

    # It is a combination of the two cells
    # One weight is the area weight and the second is the emission weight
    # weights must be scaled with the respective area of that cell used
    weigths = np.array([(1.0 / 8.0) * (3.0 / 7.0), (3.0 / 8.0) * (4.0 / 7.0)])
    # As these are the only weights used, the sum of the weights should be 1
    weigths = weigths / weigths.sum()
    expected_profile = (
        profiles.ratios[index_in] * weigths[0] + profiles.ratios[index_in2] * weigths[1]
    )

    np.testing.assert_almost_equal(
        new_profiles.ratios[index_out2],
        expected_profile,
        err_msg="The profiles should be the same",
    )
